/** \file fifo.c
 *
 * Stupid software FIFO for test purposes.
 *
 * (c) 2009 Martin Strubel <hackfin@section5.ch>
 *
 * Changes:
 *  09/2011 Martin Strubel <hackfin@section5.ch>
 *     Implemented a first-fall-through FIFO for the GHDL simulator
 *     interface.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "fifo.h"
#include "ghpi.h"

extern Fifo g_fifos[2];

int fifo_init(Fifo *f, unsigned short size)
{
	f->size = size;
	f->buf = (unsigned char*) malloc(size);
	if (!f->buf) return -1;
	f->head = 0;
	f->tail = 0;
	f->fill = 0;
	f->ovr = LOW;
	f->unr = LOW;
	return 0;
}

void fifo_exit(Fifo *f)
{
	free(f->buf);
}

int fifo_getbyte(Fifo *f, unsigned char *byte)
{
	if (f->tail == f->head) {
		return 0;
	}
	*byte = f->buf[f->tail];
	return 1;
}

int fifo_advance(Fifo *f)
{
	if (f->fill == 0) {
		f->unr = HIGH;
		fprintf(stderr, "Error: FIFO underrun\n");
		return 0;
	}
	f->tail++; f->tail %= f->size;
	f->fill--;

	return 1;
}

int fifo_read(Fifo *f, unsigned char *byte, unsigned short n)
{
	int i = 0;
	if (n == 0) return 0;

	if (f->fill == 0) {
		f->unr = HIGH;
		fprintf(stderr, "Error: FIFO underrun\n");
		return 0;
	}
	do {
		*byte++ = f->buf[f->tail++];
		f->tail %= f->size;
		n--; i++;
	} while (f->tail != f->head && n > 0);
	f->fill -= i;

	return i;
}

int fifo_fill(Fifo *f)
{
	return f->fill;
}

int fifo_status(Fifo *f, char which)
{
	if (which) { // Read?
		if (f->fill == 0) return 0;
	} else {
		if (f->fill == f->size) return 0;
	}
	return 1;
}

void fifo_reset(Fifo *f)
{
	f->unr = LOW;
	f->ovr = LOW;
}


int fifo_write(Fifo *f, const unsigned char *byte, unsigned short n)
{
	int i = 0;

	if (n == 0) return 0;

	if (f->fill == f->size) {
		f->ovr = HIGH;
		fprintf(stderr, "Error: FIFO overrun\n");
		return 0;
	}

	do {
		f->buf[f->head++] = *byte++;
		f->head %= f->size;
		n--; i++;
	} while (f->tail != f->head && n > 0);

	f->fill += i;

	return i;
}

/** This is the FIFO filler/emptier
 * 
 * ..for the FALLTHROUGH FIFO type.
 *
 * \param flag   0: read (set out), 1: write (get in)
 *
 *
 */

void sim_fifo_io(char *data, char *flag)
{
// 	printf("in: %p, out: %p, flag: %x\n", in, out, flag);
	uint32_t v;
	unsigned char b;
	unsigned char rx_available;

	// Check W1C error flags:
	if (flag[OVR] == HIGH) g_fifos[TO_SIM].unr = LOW;
	if (flag[UNR] == HIGH) g_fifos[FROM_SIM].ovr = LOW;

	// Do we write?
	if (flag[FROM_SIM] == HIGH) {
		logic_to_uint(data, 8, &v);
		b = v;
		fifo_write(&g_fifos[FROM_SIM], &b, 1);
		// Feed back status SIM->FIFO queue
	}

	// Did we read advance?
	if (flag[TO_SIM] == HIGH) {
		fifo_advance(&g_fifos[TO_SIM]);
	}

	// See if byte available:
	rx_available = fifo_getbyte(&g_fifos[TO_SIM], &b);
	if (rx_available) {
		uint_to_logic(data, 8, b);
		flag[TO_SIM] = HIGH; // we have data ready
	} else {
		flag[TO_SIM] = LOW;
	}


	flag[TO_SIM]   = fifo_status(&g_fifos[TO_SIM]  , FIFO_READ ) ? HIGH : LOW;
	flag[FROM_SIM] = fifo_status(&g_fifos[FROM_SIM], FIFO_WRITE) ? HIGH : LOW;

	// Return OVR/UNR flags
	flag[OVR] = g_fifos[FROM_SIM].ovr;
	flag[UNR] = g_fifos[TO_SIM].unr;
}

// not first fall through
void sim_fifo_io_ex(char *data, char *flag)
{
	uint32_t v;
	unsigned char b;


	// Check W1C error flags:
	if (flag[OVR] == HIGH) g_fifos[TO_SIM].unr = LOW;
	if (flag[UNR] == HIGH) g_fifos[FROM_SIM].ovr = LOW;

	// Are we serious to operate?
	if (flag[TO_SIM] == HIGH) {
		// printf("operate\n");
		if (fifo_read(&g_fifos[TO_SIM], &b, 1) == 1) {
			uint_to_logic(data, 8, b);
		}
	}

	if (flag[FROM_SIM] == HIGH) {
		logic_to_uint(data, 8, &v);
		b = v;
		fifo_write(&g_fifos[FROM_SIM], &b, 1);
	}

	// Feed back status of FIFO->SIM queue
	if (fifo_status(&g_fifos[TO_SIM], FIFO_READ)) {
		flag[TO_SIM] = HIGH; // we have data ready
	} else {
		flag[TO_SIM] = LOW;
	}
	
	if (fifo_status(&g_fifos[FROM_SIM], FIFO_WRITE)) {
		flag[FROM_SIM] = HIGH; // clear to send
	} else {
		flag[FROM_SIM] = LOW;
	}

	// Return OVR/UNR flags
	flag[OVR] = g_fifos[FROM_SIM].ovr;
	flag[UNR] = g_fifos[TO_SIM].unr;
}
