/** \file thread.c
 *
 * Thread demo for GHDL C simulator interface
 * (c) 2011, Martin Strubel <hackfin@section5.ch>
 *
 * Compile with -DUSE_NETPP if you want to use the netpp server
 * interface.
 *
 */

#include <pthread.h>
#include <stdio.h>
#include <unistd.h>

#include "ghpi.h"
#include "fifo.h"

#ifdef USE_NETPP
#include "slave.h"
#include "netpp.h"
#endif

#define FIFO_SIZE   512

int fifo_blocking_read(Fifo *f, unsigned char *buf, unsigned int n)
{
	int i;
	int retry = 5;

	while (n) {
		while (!(fifo_status(f, FIFO_READ))) {
			usleep(100000);
			retry--;
			if (retry == 0) return -1;
		}
		i = fifo_read(f, buf, n);
		buf += i; n -= i;
	}
	return n;
}

#ifdef USE_NETPP

int init_backend(void)
{
	register_proplist(g_devices, g_ndevices);
	return 0;
}


void *fifo_thread(void *arg)
{
	int error;
	char *argv[] = {
		"",
	};
	error = init_backend();
	if (error < 0) return 0;
	error = start_server(1, argv);
	if (error < 0) return 0;
	return (void *) 1;
}

#else

void *fifo_thread(void *arg)
{
	int n;
	int error;

	static
	unsigned char buf[FIFO_SIZE];

	static
	unsigned char seq[] = "Don't you wanna know what's cool?";

	int i = 10;

	Fifo *fifos = (Fifo *) arg;

	while (i--) {
		usleep(1000);
		if (fifo_status(&fifos[TO_SIM], FIFO_WRITE)) {
			n = fifo_write(&fifos[TO_SIM], seq, sizeof(seq));
			if (n > 0) {
				error = fifo_blocking_read(&fifos[FROM_SIM], buf, n);
				if (error < 0) {
					printf("Timed out\n");
				} else {
					printf("Return %d bytes from Simulator:\n", n);
					printf("%s\n", buf);
					// hexdump((char *) buf, n);
				}
			}
		} else {
			printf("FIFO to Sim not ready. Skipping.\n");
		}
		usleep(200000);
	}
	// Send TERMINATE command:
	n = fifo_write(&fifos[TO_SIM], (unsigned char *) "\377", 1);

	return 0;
}


#endif

pthread_t g_thread;
Fifo g_fifos[2];

int thread_init()
{
	int error;

	error = fifo_init(&g_fifos[TO_SIM], FIFO_SIZE);
	if (error < 0) return error;
	error = fifo_init(&g_fifos[FROM_SIM], FIFO_SIZE);
	if (error < 0) return error;

	error = pthread_create(&g_thread, NULL, &fifo_thread, g_fifos);
	if (error < 0) return error;
	return 0;
}


void thread_exit()
{
	int error;
	error = pthread_cancel(g_thread);
	fifo_exit(&g_fifos[0]);
	fifo_exit(&g_fifos[1]);
}
